import json
import sys
import os

from evaluation import start, build_evaluation, write_metrics_result



def read_config(config_file_name):

    with open(config_file_name, "r") as fd:
        config_exp = json.load(fd)
    return config_exp

def run():

    config_file_name = sys.argv[1]

    config_exp = read_config(config_file_name)

    for name_experiment in config_exp:

        if not(os.path.exists(name_experiment)):
            os.mkdir(name_experiment)

        language = config_exp[name_experiment]["language"]
        start(language)
        tools = config_exp[name_experiment]["tools"]

        narrative_elements = {"participant":tools["participant"],\
                          "time":tools["time"],\
                          "event":tools["event"],\
                          "srlink":tools["srlink"]}

        data_dir = config_exp[name_experiment]["data"]
        res = build_evaluation(narrative_elements=narrative_elements, language=language, data_dir=data_dir,
                           results_dir=name_experiment)

        with open("%s_metrics.txt" % name_experiment, "w") as fd:
            write_metrics_result(res, fd)


if __name__ == "__main__":
    run()



from typing import DefaultDict, List, Optional, Iterator, Set, Tuple
from collections import defaultdict
import codecs
import os
import logging


class Preprocess():
    """ 
    Preprocess the input data according to the steps described in Daniel Falci's thesis
    Inputs:
    Parameters
    ----------
    tokens : ``List[str]``
        List of instance tokens.
    tags: ``List[str]``, (default = None)
        List of the associated tags for this instance. 
        When predicting, this field isn't used.
    separate : ``bool``, (default = True)
        Whether to separate the words joined by "_".   
    contract : ``bool``, (default = True)
        Whether to perform word contractions (e.g.: "de"+"o" -> "do")

    """
    def __init__(self,
                 tokens: List[str],
                 tags: List[str] = None,
                 separate: bool = True, 
                 contract: bool = True
    ) -> None:
        self.tokens = tokens
        self.tags = tags
        self.has_tags = tags!=None
        self.art_def = ['a', 'as', 'o', 'os']
        self.art_def_masc = ['o', 'os']
        self.adverb = ['aí', 'aqui', 'ali'] 
        self.pron = ['ele', 'eles', 'ela', 'elas', 'esse', 'esses', 'essa', 'essas', 'isso', 'este', 'estes', 'esta', 'estas', 'isto'] 
        self.pron2 = ['aquele', 'aqueles', 'aquela', 'aquelas', 'aquilo'] 
        self.ref = ['o', 'os', 'a', 'as', 'me', 'se', 'te', 'vos', 'lhe', 'lho', 'lhas', 'lhos', 'lha', 'lo', 'la', 'los', 'las', 'lhes', 'no', 'na', 'nos']
        self.change = self.art_def + self.adverb + self.pron + self.pron2 + self.ref
        self.prepositions = ['de', 'a', 'em', 'por', '-']
        self._separate = separate
        self._contract = contract

    def preprocess(self):
        if self._separate:
            self.separate()
        if self._contract:
            self.contractions()
        #Check that tokens and tags list has same size
        if self.has_tags and len(self.tokens) != len(self.tags):
            raise Exception("Lengths of token list %i and tag list %i don't match." % (len(self.tokens), len(self.tags)))
        return self.tokens, self.tags

    def separate(self):
        """
        We want to separate multiword nouns and expressions such as 
        "Campeonato_Brasileiro" and "em_termos_de" into several tokens
        (removing the "_")
        """
        ind = 0
        while ind < len(self.tokens):
            if '_' in self.tokens[ind] and '_' != self.tokens[ind]: #sometimes '_' represents '--'
                new_tokens = self.tokens[ind].split('_')
                self.tokens = self.tokens[:ind] + new_tokens + self.tokens[ind+1:]
                if self.has_tags:
                    if self.tags[ind][:2] == 'B-':
                        self.tags = self.tags[:ind+1] + ['I-' + self.tags[ind][2:]]*(len(new_tokens)-1) + self.tags[ind+1:]
                    else:
                        self.tags = self.tags[:ind] + [self.tags[ind]]*len(new_tokens) + self.tags[ind+1:] 
            ind += 1

    def determine_tags(self, ind, new_token):
        """
        Given a new, contracted token at position ind, 
        change tags list to reflect the new tokens list.
        """
        if self.tags[ind] == self.tags[ind+1]:
            del self.tags[ind]
        elif self.tags[ind] == 'O' and self.tags[ind+1] != 'O':
            del self.tags[ind]
        elif self.tags[ind] != 'O' and self.tags[ind+1] == 'O':
            del self.tags[ind+1]
            if "C-" in self.tags[ind+1] and self.tags[ind+1][4:] == self.tags[ind][2:]:
                for i, tag in enumerate(self.tags[ind+1:],ind+1):
                    if "C-" + self.tags[ind][2:] in tag:
                        self.tags[i] = "I-" + tag[4:]
        elif self.tags[ind][:2] == "B-" and self.tags[ind+1][:2] == "I-" and self.tags[ind][2:] == self.tags[ind+1][2:]:
            del self.tags[ind+1]
        else:
            raise Exception()

    def contractions(self):
        """
        We want to contract prepositions and articles, pronouns and adverbs
        These contractions are widely used in the portuguese language and a lot of times
        are used instead of the separated form even in formal settings
        """
        if any(token in self.change for token in self.tokens) and any(token.lower() in self.prepositions for token in self.tokens):
            ind = 0
            while ind < len(self.tokens)-1:
                # The only captalization that matters is in the first word of the contraction
                is_cap = self.tokens[ind][0].isupper()

                prev_token = self.tokens[ind].lower()
                cur_token = self.tokens[ind+1].lower()

                if prev_token in self.prepositions and cur_token in self.change:
                    if prev_token == 'de' and cur_token in self.art_def + self.adverb + self.pron + self.pron2: 
                        new_token = 'd' + cur_token
                    elif prev_token == 'em' and cur_token in self.art_def + self.pron + self.pron2:
                        new_token = 'n' + cur_token
                    elif prev_token == 'a' and cur_token in self.art_def + self.pron2 and cur_token not in self.art_def_masc:
                        new_token = 'à' + cur_token[1:]
                    elif prev_token == 'a' and cur_token in self.art_def_masc:
                        new_token = 'a' + cur_token
                    elif prev_token == 'por' and cur_token in self.art_def:
                        new_token = 'pel' + cur_token
                    elif prev_token == '-' and cur_token in self.ref:
                        new_token = '-' + cur_token
                    else:
                        ind += 1
                        continue
                    
                    if is_cap:
                        new_token = new_token.capitalize()
                    self.tokens = self.tokens[:ind] + [new_token] + self.tokens[ind+2:]
        
                    if self.has_tags:
                        self.determine_tags(ind, new_token)
                ind += 1
                



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.14.5
# ---

# author: evelin amorim
# tries to learn the concept of participant from the result of an SRL output
import os.path
from builtins import zip

import torch

from text2story.annotators import ALLENNLP
from text2story.readers import read_brat
from text2story.core.utils import bsearch_tuplelist

from transformers import BertTokenizer, BertModel
from torch.optim import Adam
from torch.utils.data import TensorDataset, DataLoader,RandomSampler

import os
from pathlib import Path

import spacy
from torch import nn

from sklearn.metrics import f1_score, confusion_matrix


class ParticipantsModel(nn.Module):
    def __init__(self, bert, hidden_size):
        super(ParticipantsModel, self).__init__()
        self.bert = bert

        # Define your layers
        self.conv1d_span = nn.Conv1d(in_channels=self.bert.config.hidden_size, out_channels=hidden_size, kernel_size=3)
        self.conv1d_sent = nn.Conv1d(in_channels=self.bert.config.hidden_size, out_channels=64, kernel_size=3,
                                     stride=1)

        self.lstm = nn.LSTM(input_size=64, hidden_size=hidden_size, num_layers=1, batch_first=True)
        self.fc = nn.Linear(in_features=hidden_size, out_features=2)
        self.sm = nn.Softmax(dim=1)

    def forward(self, sent, span):
        sent_embeddings = self.bert(sent)[0]
        span_embeddings = self.bert(span)[0]

        # Apply 1D convolution to span_embeddings
        span_features = self.conv1d_span(span_embeddings.transpose(1, 2))

        # Apply the convolutional layer
        sentence_features = self.conv1d_sent(
            sent_embeddings.transpose(1, 2))  # [batch_size, out_channels, sequence_length - kernel_size + 1]

        # Apply the LSTM layer
        lstm_output, _ = self.lstm(
            sentence_features.transpose(1, 2))  # [batch_size, sequence_length - kernel_size + 1, hidden_size]
        # Concatenate the span embeddings with the output of the LSTM layer
        combined_embeddings = torch.cat((span_embeddings.transpose(1, 2), lstm_output),
                                        dim=1)  # [batch_size, span_length + sequence_length - kernel_size + 1, hidden_size]

        # Apply the fully connected layer
        logits = self.fc(combined_embeddings)  # [batch_size, span_length + sequence_length - kernel_size + 1, 2]
        # print("Logits size after fc: ", logits.size())
        # Apply softmax activation if needed
        logits = self.sm(logits)
        # print("Logits size after sm: ", logits.size())
        logits = torch.max(logits, dim=1)[0]
        # print("Logits size after torch.max 1: ", logits.size())
        logits = torch.max(logits, dim=1)[0]
        # print("Logits size after torch.max 2: ", logits.size())

        return logits


class ParticipantConceptLearning:

    def __init__(self, lang):
        self.lang = lang
        if lang == "en":
            if not (spacy.util.is_package('en_core_web_lg')):
                spacy.cli.download('en_core_web_lg')

            self.language_model = spacy.load('en_core_web_lg')
        if lang == "pt":
            if not (spacy.util.is_package('pt_core_news_lg')):
                spacy.cli.download('pt_core_news_lg')
            self.language_model = spacy.load('pt_core_news_lg')

        # Load the pre-trained BERT tokenizer and model
        self.max_seq_len = 128
        self.tokenizer = BertTokenizer.from_pretrained('neuralmind/bert-base-portuguese-cased',
                                                       model_max_length=self.max_seq_len)
        self.bert_model = BertModel.from_pretrained('neuralmind/bert-base-portuguese-cased')

        self.hidden_size = 128
        self.nepochs = 6
        self.learning_rate = 0.00005

        self.threshold = 0.5

        # Define the optimizer
        self.model = ParticipantsModel(self.bert_model, self.hidden_size)
        self.optimizer = Adam(self.model.parameters(), lr=self.learning_rate)
        # Define the loss function
        self.loss_fn = nn.BCELoss()

    def read_file_lst(self, file_lst):
        doc_lst = []

        for f in file_lst:
            p = Path(f)
            reader = read_brat.ReadBrat()

            file_name = p.parent.joinpath(p.stem).absolute()

            doc = reader.process_file(str(file_name))
            doc_lst.append((doc, str(p.absolute())))
        return doc_lst

    def load_dataset(self, data_dir, type_data="train", split_data=None):
        if split_data is None:
            file_lst = []
            for dirpath, dirnames, filenames in os.walk(data_dir):
                for f in filenames:
                    if f.endswith(".txt"):
                        file_lst.append(os.path.join(dirpath, f))
        else:
            if type_data == "train":
                split_data_file = os.path.join(split_data, "train.txt")
            else:
                split_data_file = os.path.join(split_data, "test.txt")

            file_lst = open(split_data_file, "r").readlines()
            file_lst = [os.path.join(data_dir, f.replace("\n", "")) for f in file_lst]

        dataset = self.read_file_lst(file_lst)

        return dataset

    def get_actors(self, doc):

        actor_lst = []

        for tok in doc:
            for ann_type, ann_attr in tok.attr:
                if ann_type == "Participant":
                    actor_lst.append((tok.offset, tok.offset + len(tok.text)))

        return actor_lst

    def get_sent_lst(self, doc_text):
        sent_lst = []
        doc_lm = self.language_model(doc_text)

        for sent in doc_lm.sents:
            sent_lst.append(sent.text)

        return sent_lst

    def get_sent_text(self, span_offset, doc_lm):
        pass

    def get_sents_offset(self, sent_lst):
        sent_offset_lst = []
        offset_start = 0
        for sent in sent_lst:
            offset_end = offset_start + len(sent) - 1
            sent_offset_lst.append((offset_start, offset_end))
            offset_start = offset_end
        return sent_offset_lst

    def get_candidates_class(self, candidate_lst, actor_lst, sent_lst):
        """

        @param sent:
        @return: a list of words
        """
        sent_off_lst = self.get_sents_offset(sent_lst)
        y = []
        sent_id_lst = []

        count_positive = 0

        for candidate in candidate_lst:

            cand_offset = candidate[0]
            sent_id = bsearch_tuplelist(cand_offset[0], sent_off_lst)
            sent_id_lst.append(sent_id)

            pos = bsearch_tuplelist(cand_offset[0], actor_lst)
            if pos == -1:
                pos = bsearch_tuplelist(cand_offset[1], actor_lst)
                if pos == -1:
                    y.append(0)
                else:
                    y.append(1)
                    count_positive += 1
            else:
                y.append(1)

        # print("Positive instances: ", count_positive / len(y))
        return candidate_lst, y, sent_id_lst

    def build_batch_data(self, dataset):

        sent_text_lst = []
        candidate_text_lst = []
        y_lst = []

        for doc, file_name in dataset:
            # I need to get the class from the lusa annotations
            with open(file_name, "r") as fd:
                doc_text = fd.read()
                candidate_lst = ALLENNLP.extract_actors(self.lang, doc_text)

            sent_lst = self.get_sent_lst(doc_text)
            actor_lst = self.get_actors(doc)

            # TODO: verificar quantidade de candidatos e quantidade de cada classe
            candidate_lst, y, sent_id_lst = self.get_candidates_class(candidate_lst, actor_lst, sent_lst)
            y_lst += y

            sent_text_lst += [sent_lst[id] for id in sent_id_lst]
            idx = 0

            for ((start, end), _, _) in candidate_lst:
                cand = doc_text[start:end]
                candidate_text_lst.append(cand)
                idx += 1
        # print("-->",len(candidate_text_lst))
        tokens_sent = self.tokenizer.batch_encode_plus(sent_text_lst,
                                                       max_length=self.max_seq_len,
                                                       pad_to_max_length=True,
                                                       truncation=True,
                                                       return_token_type_ids=False)
        tokens_sent = torch.tensor(tokens_sent['input_ids'])

        tokens_span = self.tokenizer.batch_encode_plus(candidate_text_lst,
                                                       max_length=self.max_seq_len,
                                                       pad_to_max_length=True,
                                                       truncation=True,
                                                       return_token_type_ids=False)
        tokens_span = torch.tensor(tokens_span['input_ids'])

        labels = torch.tensor(y_lst)

        mydataset = TensorDataset(tokens_sent, labels, tokens_span)
        mydataloader = DataLoader(mydataset, batch_size=32, shuffle=True)

        return mydataloader

    def process_train(self, data_dir, device, split_dir=None):

        self.model.train()

        # load train dataset
        if split_dir is not None:
            dataset = self.load_dataset(data_dir, type_data="train", split_data=split_dir)
        else:
            dataset = self.load_dataset(data_dir)

        # load srl model
        ALLENNLP.load(self.lang)

        # apply srl and get possible candidates as participants
        dataloader = self.build_batch_data(dataset)
        self.model = self.model.to(device)

        # empty list to save model predictions
        total_preds = []
        total_labels = []

        total_loss = 0.0
        total_accuracy = 0.0

        torch.manual_seed(2023)

        # iterate over batches
        for step, (batch, labels, span) in enumerate(dataloader):
            if step % 10 == 0:
                print(" STEP ", step)

            sent = batch.to(device)
            span = span.to(device)

            # clear previously calculated gradients
            self.model.zero_grad()

            # get model predictions for the current batch
            # Forward pass
            outputs = self.model(sent, span)

            # Calculate the loss
            labels = labels.to(device).float()

            loss = self.loss_fn(outputs, labels)
            total_loss += loss.item()

            outputs = (outputs >= self.threshold).int()

            # Backpropagation
            loss.backward()

            # clip the the gradients to 1.0. It helps in preventing the exploding gradient problem
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)

            # update paremeters
            self.optimizer.step()

            # model predictions are stored on GPU. So, push it to CPU
            outputs = outputs.detach().cpu().numpy()

            # outputs = np.argmax(outputs, axis=1)

            # append the model predictions
            total_preds += outputs.tolist()
            total_labels += labels.tolist()

        # Print average loss for the epoch
        avg_loss = total_loss / len(dataloader)
        # print(f"Epoch {epoch+1}/{self.nepochs}, Average Loss: {avg_loss}")

        # predictions are in the form of (no. of batches, size of batch, no. of classes).
        # reshape the predictions in form of (number of samples, no. of classes)
        # total_preds  = np.concatenate(total_preds, axis=0)
        f1 = f1_score(total_labels, total_preds, average='weighted')
        m = confusion_matrix(total_labels, total_preds)

        print("\nCONFUSION MATRIX")
        print(m)
        print()

        # returns the loss and predictions
        return avg_loss, f1
        # return 0,0
        # function for evaluating the model

    def evaluate(self, data_dir, device, split_dir=None):

        print("\nEvaluating...")

        # deactivate dropout layers
        self.model.eval()
        # Define the loss function
        loss_fn = nn.BCELoss()

        total_loss, total_accuracy = 0, 0

        # empty list to save the model predictions
        total_preds = []
        total_labels = []

        # load validation dataset
        if split_dir is not None:
            dataset = self.load_dataset(data_dir, type_data="test", split_data=split_dir)
        else:
            dataset = self.load_dataset(data_dir)

        # apply srl and get possible candidates as participants
        dataloader = self.build_batch_data(dataset)
        self.model = self.model.to(device)

        # iterate over batches
        for step, (batch, labels, span) in enumerate(dataloader):

            # Progress update every 10 batches.
            if step % 5 == 0 and not step == 0:
                # Calculate elapsed time in minutes.
                # elapsed = format_time(time.time() - t0)

                # Report progress.
                print('  Batch {:>5,}  of  {:>5,}.'.format(step, len(dataloader)))

            # push the batch to gpu
            sent = batch.to(device)
            span = span.to(device)

            # deactivate autograd
            with torch.no_grad():

                # model predictions
                preds = self.model(sent, span)

                # compute the validation loss between actual and predicted values
                labels = labels.to(device).float()
                loss = self.loss_fn(preds, labels)

                total_loss = total_loss + loss.item()

                preds = (preds >= self.threshold).int()
                preds = preds.detach().cpu().numpy()
                # preds = np.argmax(preds, axis=1)
                total_preds += list(preds)
                total_labels += labels.tolist()

        # compute the validation loss of the epoch
        avg_loss = total_loss / len(dataloader)

        # reshape the predictions in form of (number of samples, no. of classes)
        # total_preds  = np.concatenate(total_preds, axis=0)

        f1 = f1_score(total_labels, total_preds, average='weighted')
        m = confusion_matrix(total_labels, total_preds)

        print("\nCONFUSION MATRIX")
        print(m)
        print()

        return avg_loss, f1







from typing import DefaultDict, List, Optional, Iterator, Set, Tuple
from collections import defaultdict
import codecs
import os
import logging
import random

from nltk import Tree

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# This file is similar to ontonotes.py
# in allennlp_models/common, but reads the portuguese
# CoNLL files (some columns are different from ontonotes)

class PropBankBrSentence:
    """
    A class representing the annotations available for a single PropBank.Br formatted sentence.

    Parameters
    ----------
    words : ``List[str]``
        This is the tokens as segmented/tokenized in the Treebank.
    predicate_lemmas : ``List[Optional[str]]``
        The predicate lemma of the words for which we have semantic role
        information or word sense information. All other indices are ``None``.
    predicate_senses : ``List[Optional[int]]``
        The VerbNet PT frameset ID of the lemmas in ``predicate_lemmas``, or ``None``.
    srl_frames : ``List[Tuple[str, List[str]]]``
        A dictionary keyed by the verb in the sentence for the given
        Propbank frame labels, in a BIO format.
    """
    def __init__(self,
                 words: List[str],
                 predicate_senses: List[Optional[str]],
                 predicate_lemmas: List[Optional[str]],
                 srl_frames: List[Tuple[str, List[str]]]) -> None:

        self.words = words
        self.predicate_senses = predicate_senses
        self.predicate_lemmas = predicate_lemmas
        self.srl_frames = srl_frames


class PropBankBr:
    """
    This DatasetReader is designed to read in the PropBank-Br v1.1 data. 

    The file path provided to this class can then be any of the train, test or development
    directories.

    The important columns are:
    1 Word : ``str``
        This is the token as segmented/tokenized in the Treebank. 
    8 Predicate Frameset ID: ``int``
        The PropBank frameset ID of the predicate in Column 7.
    9 Predicate lemma: ``str``
        The predicate lemma is mentioned for the rows for which we have semantic role
        information or word sense information. All other rows are marked with a "-".
    10+ Predicate Arguments: ``str``
        There is one column each of predicate argument structure information for the predicate
        mentioned in Column 7. If there are no predicates tagged in a sentence this is a
        single column with all rows marked with an ``*``.
    """
    def __init__(self, remove_c):
        self.flag = []
        # Flag to remove the continuation arguments (to compare with Falci et al.):
        self.remove_c = remove_c

    def dataset_iterator(self, file_path: str) -> Iterator[PropBankBrSentence]:
        """
        An iterator over the entire dataset, yielding all sentences processed.
        """
                
        for sentence in self.dataset_document_iterator(file_path):
            yield sentence

    def dataset_document_iterator(self, file_path: str) -> List[PropBankBrSentence]:
        """
        An iterator over CONLL formatted files which yields documents, regardless
        of the number of document annotations in a particular file. This is useful
        for conll data which has been preprocessed, such as the preprocessing which
        takes place for the 2012 CONLL Coreference Resolution task.
        """
        with codecs.open(file_path, 'r', encoding='UTF-8') as open_file:
            conll_rows = []
            document: List[PropBankBrSentence] = []
            for line in open_file:
                line = line.strip()
                if line != '' and not line.startswith('#'):
                    # Non-empty line. Collect the annotation.
                    conll_rows.append(line)
                else:
                    if conll_rows:
                        sentence = self._conll_rows_to_sentence(conll_rows)
                        if len(self.flag) != 0: 
                            # Remove instances that have either C- tags 
                            # (if we want to remove them) or double labels
                            sentence.srl_frames = [srl_frame for ind, srl_frame in enumerate(sentence.srl_frames) if ind not in self.flag]
                            self.flag = []
                            if len(sentence.srl_frames) == 0:
                                # If there are no annotated verbs left
                                conll_rows = []
                                continue
                        document.append(sentence)
                        conll_rows = []
                if line.startswith("#end document"):
                    return document
            if document:
                # Collect any stragglers or files which might not
                # have the '#end document' format for the end of the file.
                return document


    def _conll_rows_to_sentence(self, conll_rows: List[str]) -> PropBankBrSentence:

        # The words in the sentence.
        sentence: List[str] = []
        # The VerboBrasil ID of the predicate.
        predicate_senses: List[str] = []
        # The lemmatised form of the predicates in the sentence 
        predicate_lemmas: List[str] = []

        verbal_predicates: List[str] = []
        span_labels: List[List[str]] = []
        current_span_labels: List[str] = []

        for _, row in enumerate(conll_rows):
            conll_components = row.split()

            word = conll_components[1]
            lemmatised_predicate = conll_components[9]
            sense = conll_components[8]

            if not span_labels:
                # If this is the first word in the sentence, create
                # empty lists to collect the SRL BIO labels.
                # We can't do this upfront, because we don't know how many
                # components we are collecting, as a sentence can have
                # variable numbers of SRL frames.
                span_labels = [[] for _ in conll_components[10:]]
                # Create variables representing the current label for each label
                # sequence we are collecting.
                current_span_labels = [None for _ in conll_components[10:]]

            self._process_span_annotations_for_word(self,
                                                    conll_components[10:],
                                                    span_labels,
                                                    current_span_labels)

            # If any annotation marks this word as a verb predicate,
            # we need to record its index. This also has the side effect
            # of ordering the verbal predicates by their location in the
            # sentence, automatically aligning them with the annotations.
            # !! perhaps change this to only adding if it has a predicate sense
            word_is_verbal_predicate = any(["(V" in x for x in conll_components[10:]])
            if word_is_verbal_predicate:
                verbal_predicates.append(word)

            sentence.append(word)
            predicate_lemmas.append(lemmatised_predicate if lemmatised_predicate != "-" else None)
            predicate_senses.append(sense if sense != "-" else None)

        srl_frames = [(predicate, labels) for predicate, labels
                      in zip(verbal_predicates, span_labels)]
        return PropBankBrSentence(sentence,
                                 predicate_senses,
                                 predicate_lemmas,
                                 srl_frames)

    @staticmethod
    def _process_span_annotations_for_word(self,
                                           annotations: List[str],
                                           span_labels: List[List[str]],
                                           current_span_labels: List[Optional[str]]) -> None:
        """
        Given a sequence of different label types for a single word and the current
        span label we are inside, compute the BIO tag for each label and append to a list.

        Parameters
        ----------
        annotations: ``List[str]``
            A list of labels to compute BIO tags for.
        span_labels : ``List[List[str]]``
            A list of lists, one for each annotation, to incrementally collect
            the BIO tags for a sequence.
        current_span_labels : ``List[Optional[str]]``
            The currently open span per annotation type, or ``None`` if there is no open span.
        """
        for annotation_index, annotation in enumerate(annotations):
            # strip all bracketing information to
            # get the actual propbank label.
            label = annotation.strip("()*")
            if self.remove_c and  "C-" in label:
                self.flag.append(annotation_index)
            if "(" in annotation:
                # Entering into a span for a particular semantic role label.
                # We append the label and set the current span for this annotation.
                if current_span_labels[annotation_index] is not None:
                    # If there is already a current label but the label has a "(", 
                    # this token has 2 annotations and it is to be considered an annotation error.
                    # Do not append to the document.
                    self.flag.append(annotation_index)
                bio_label = "B-" + label
                span_labels[annotation_index].append(bio_label)
                current_span_labels[annotation_index] = label
                
            elif current_span_labels[annotation_index] is not None:
                # If there's no '(' token, but the current_span_label is not None,
                # then we are inside a span.
                bio_label = "I-" + current_span_labels[annotation_index]
                span_labels[annotation_index].append(bio_label)
            else:
                # We're outside a span.
                span_labels[annotation_index].append("O")
            # Exiting a span, so we reset the current span label for this annotation.
            if ")" in annotation:
                current_span_labels[annotation_index] = None



"""
    spaCy annotator

    Used for:
        - Actor extraction
            'pt' : https://spacy.io/models/en#en_core_web_lg)
            'en' : https://spacy.io/models/en#en_core_web_lg)
"""

from text2story.core.utils import chunknize_actors
from text2story.core.exceptions import InvalidLanguage

import spacy

pipeline = {}

def load(lang):
    """
    Used, at start, to load the pipeline for the supported languages.
    """
    if lang == "pt":
        if not(spacy.util.is_package('pt_core_news_lg')):
            spacy.cli.download('pt_core_news_lg')
        pipeline['pt'] = spacy.load('pt_core_news_lg')

    elif lang == "en":

        if not(spacy.util.is_package('en_core_web_lg')):
            spacy.cli.download('en_core_web_lg')

        pipeline['en'] = spacy.load('en_core_web_lg')
    else:
        raise InvalidLanguage(lang)

    
def extract_actors(lang, text):
    """
    Parameters
    ----------
    lang : str
        the language of text to be annotated
    text : str
        the text to be annotated
    
    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        the list of actors identified where each actor is represented by a tuple

    Raises
    ------
        InvalidLanguage if the language given is invalid/unsupported
    """

    if lang not in ['pt', 'en']:
        raise InvalidLanguage(lang)

    doc = pipeline[lang](text)

    iob_token_list = []
    for token in doc:
        start_character_offset = token.idx
        end_character_offset = token.idx + len(token)
        character_span = (start_character_offset, end_character_offset)
        pos = normalize(token.pos_)
        ne = token.ent_iob_ + "-" + normalize(token.ent_type_) if token.ent_iob_ != 'O' else 'O'

        iob_token_list.append((character_span, pos, ne))

    actor_list = chunknize_actors(iob_token_list)

    return actor_list  


def normalize(label):
    """
    Parameters
    ----------
    label : str
    
    Returns
    -------
    str
        the label normalized
    """

    mapping = {
        # POS tags
        # Universal POS Tags
        # http://universaldependencies.org/u/pos/
        
        #"ADJ": "adjective",
        #"ADP": "adposition",
        #"ADV": "adverb",
        #"AUX": "auxiliary",
        #"CONJ": "conjunction",
        #"CCONJ": "coordinating conjunction",
        #"DET": "determiner",
        #"INTJ": "interjection",
        "NOUN": "Noun",
        #"NUM": "numeral",
        #"PART": "particle",
        "PRON": "Pronoun",
        "PROPN": "Noun",
        #"PUNCT": "punctuation",
        #"SCONJ": "subordinating conjunction",
        #"SYM": "symbol",
        #"VERB": "verb",
        #"X": "other",
        #"EOL": "end of line",
        #"SPACE": "space",

        # NE
        # en
        'CARDINAL'    : 'Other', # 'Numerals that do not fall under another type'
        'DATE'        : 'Date',  # 'Absolute or relative dates or periods'
        'EVENT'       : 'Other', # 'Named hurricanes, battles, wars, sports events, etc.'
        'FAC'         : 'Loc',   # 'Buildings, airports, highways, bridges, etc.'
        'GPE'         : 'Loc',   # 'Countries, cities, states'
        'LANGUAGE'    : 'Other', # 'Any named language'
        'LAW'         : 'Other', # 'Named documents made into laws.'
        'LOC'         : 'Loc',   # 'Non-GPE locations, mountain ranges, bodies of water'
        'MONEY'       : 'Other', # 'Monetary values, including unit'
        'NORP'        : 'Other', # 'Nationalities or religious or political groups'
        'ORDINAL'     : 'Other', # '"first", "second", etc.'
        'ORG'         : 'Org',   # 'Companies, agencies, institutions, etc.'
        'PERCENT'     : 'Other', # 'Percentage, including "%"'
        'PERSON'      : 'Per',   # 'People, including fictional'
        'PRODUCT'     : 'Obj',   # 'Objects, vehicles, foods, etc. (not services)'
        'QUANTITY'    : 'Other', # 'Measurements, as of weight or distance'
        'TIME'        : 'Time',  # 'Times smaller than a day'
        'WORK_OF_ART' : 'Other', # 'Titles of books, songs, etc.'

        # pt
        # 'LOC'
        'MISC'        : 'Other', # 'Miscellaneous entities, e.g. events, nationalities, products or works of art'
        # 'ORG'
        'PER'         : 'Per' # 'People, including fictional'
    }

    return mapping.get(label, 'UNDEF')
        



import sys
from pathlib import Path

from text2story.core.exceptions import InvalidTool
from text2story.annotators import SPACY, NLTK, SPARKNLP
from text2story.annotators import PY_HEIDELTIME, ALLENNLP, CUSTOMPT
from text2story.annotators import SRLWeakLabeling

sys.path.insert(0, Path(__file__).parent)

ACTOR_EXTRACTION_TOOLS = ['spacy', 'nltk', 'sparknlp','allennlp','srlweaklabeling']
TIME_EXTRACTION_TOOLS = ['py_heideltime']
EVENT_EXTRACTION_TOOLS = ['allennlp','custompt']
OBJECTAL_LINKS_RESOLUTION_TOOLS = ['allennlp']
SEMANTIC_ROLE_LABELLING_TOOLS = ['allennlp']


def load(lang):
    SPACY.load(lang)
    NLTK.load(lang)
    #SPARKNLP.load(lang)
    PY_HEIDELTIME.load(lang)
    ALLENNLP.load(lang)
    CUSTOMPT.load(lang)

def get_available_tools():
    return ACTOR_EXTRACTION_TOOLS+\
            TIME_EXTRACTION_TOOLS+\
            EVENT_EXTRACTION_TOOLS+\
            OBJECTAL_LINKS_RESOLUTION_TOOLS+\
            SEMANTIC_ROLE_LABELLING_TOOLS

def get_participant_tools():
    return ACTOR_EXTRACTION_TOOLS

def get_time_tools():
    return TIME_EXTRACTION_TOOLS

def get_event_tools():
    return EVENT_EXTRACTION_TOOLS

def get_srlink_tools():
    return SEMANTIC_ROLE_LABELLING_TOOLS

def extract_actors(tool, lang, text):

    if tool == 'spacy':
        return SPACY.extract_actors(lang, text)
    elif tool == 'nltk':
        return NLTK.extract_actors(lang, text)
    elif tool == 'allennlp':
        return ALLENNLP.extract_actors(lang, text)
    elif tool == 'srlweaklabeling':
        return SRLWeakLabeling.extract_actors(lang, text)
    #elif tool == 'sparknlp':
    #    return SPARKNLP.extract_actors(lang, text)

    raise InvalidTool


def extract_times(tool, lang, text, publication_time):
    if tool == 'py_heideltime':
        return PY_HEIDELTIME.extract_times(lang, text, publication_time)

    raise InvalidTool


def extract_objectal_links(tool, lang, text):
    if tool == 'allennlp':
        return ALLENNLP.extract_objectal_links(lang, text)

    raise InvalidTool


def extract_events(tool, lang, text):
    if tool == 'allennlp':
        return ALLENNLP.extract_events(lang, text)
    if tool == 'custompt':
        return CUSTOMPT.extract_events(lang, text)

    raise InvalidTool


def extract_semantic_role_links(tool, lang, text):
    if tool == 'allennlp':
        return ALLENNLP.extract_semantic_role_links(lang, text)

    raise InvalidTool




import json

from dataclasses import dataclass, field
from typing import List

from text2story.readers.token_corpus import TokenCorpus
from text2story.select.event import get_all_events, get_event_type
from text2story.select.event import sieve_bbubbles_events, is_event

def is_type_rel(type_rel, type_rel_lst):

        for t in type_rel_lst:
            if type_rel.startswith(t):
                return True
        return False

class Bubble:

    def __init__(self, event = None):
        #print("-->", event)
        self.event_chain = []
        if isinstance(event, list):

            self.event = event[0]
            if len(event) > 1:
                self.event_chain = event
        else:
            self.event = event # event trigger as TokenCorpus
        #if event is not None:
        #    print("-->", event.id_ann)
        self.relations = [] # the edges to other bubbles (BubbleEdge)
       
        self.name = ""
        # a text span (type:[TokenCorpus]) that is connected 
        # by an SRLINK_agent to this event
        self.agent = Agent() 

    def get_event_text(self):
        if self.event_chain == []:
            return self.event.text
        else:
            tok_txt = [tok.text for tok in self.event_chain]
            return " ".join(tok_txt)

    def get_id_ann(self):
        return self.event.id_ann

    def get_hist_relations(self):

        hist_rel = {}

        for bubble_rel in self.relations:
            if bubble_rel.edge_type in hist_rel:
                hist_rel[bubble_rel.edge_type] += 1
            else:
                hist_rel[bubble_rel.edge_type] = 1
        return hist_rel

    def add_agent_relations(self, map_ann, relations_set, type_rel_lst = []):

        # this bubble does not have an agent object
        if len(self.agent.span) == 0:
            return


        # iterate through the relations of the agent of
        # the event in the bubble
        current_agent = self.agent
        while current_agent != None:


            for rel in current_agent.span[0].relations:

                for id_ann_rel in rel.toks[0].id_ann:

                    if rel.rel_type in type_rel_lst and rel.rel_id not in relations_set:

                        if id_ann_rel in map_ann:
                            object_rel = map_ann[id_ann_rel]


                            if isinstance(object_rel, Agent):

                                relation = RelationsAgent(object_rel,
                                                          rel.rel_type,
                                                          rel.rel_id)
                                relation.set_arg(rel.argn)

                                current_agent.relations_agent.append(relation)
                            else:
                                relation = RelationsBubbleAgent(object_rel,
                                                                rel.rel_type,
                                                                rel.rel_id)
                                relation.set_arg(rel.argn)
                                current_agent.relations_bubbles.append(relation)

                            relations_set.add(rel.rel_id)
            current_agent = current_agent.next_agent

    def add_relations(self, map_ann,  relations_set, type_rel_lst = []):



        for rel in self.event.relations:
            self_event_type = get_event_type(self.event)

            for idx_rel, id_ann_rel in enumerate(rel.toks[0].id_ann):
                ann_type_rel, attr_rel = rel.toks[0].attr[idx_rel]

                # getting the participant connected to this bubble
                if "Reporting" in self_event_type and\
                     (rel.rel_type.startswith("SRLINK")) and\
                     ann_type_rel == "Participant" and\
                     id_ann_rel == rel.arg_id and\
                     rel.toks[0].sent_id == int(self.event.sent_id):
                     # the last condition test if the id of this token chain if related
                     # to the current relation


                     agent_pointer = self.agent.add_agent(rel.toks)
                     id_ann_agent = rel.toks[0].id_ann[0]
                     map_ann[id_ann_agent] = agent_pointer

                     #print("Agent %s Event %s ID REL %s %d %s" % (id_ann_agent, self.get_id_ann(), rel.rel_id, idx_rel, rel.arg_id))

                if id_ann_rel in map_ann and\
                 rel.rel_id not in relations_set and\
                 is_type_rel(rel.rel_type, type_rel_lst):

                     if not(isinstance(map_ann[id_ann_rel], Agent)):
                         #print("Event %s ID REL %s %d %s %s" % (self.get_id_ann(), rel.rel_id, idx_rel, rel.arg_id, rel.rel_type))
                         bubble_rel = map_ann[id_ann_rel]
                         relation = BubbleRelation(rel.rel_type,\
                                              bubble_rel,\
                                              rel.argn)

                         self.relations.append(relation)
                         relations_set.add(rel.rel_id)

                    #prnt("==>", rel.rel_type)
                    #print("-->", rel.rel_id)
                    #print("-->",rel.argn)
@dataclass
class Agent:

    span: [TokenCorpus] = field(default_factory=list)
    relations_bubbles:['RelationsBubbleAgent'] = field(default_factory=list)
    relations_agent: List['RelationsAgent'] = field(default_factory=list)
    name: str = ""
    next_agent: 'Agent' = None

    def add_agent(self, toks: [TokenCorpus]):
        if self.span != []:
            current_agent = self
            while current_agent.next_agent != None:
                current_agent = current_agent.next_agent

            current_agent.next_agent = Agent(span = toks)
            return current_agent.next_agent

        else:
            self.span = toks
            return self

@dataclass
class RelationsBubbleAgent:
    bubble: Bubble
    rel_type: str
    rel_id:str

    def set_arg(self, argn):

        # if the relation starts with the agent pointer,
        # then its type is out
        if argn == "arg1":
            self.out = True
        else:
            self.out = False
@dataclass
class RelationsAgent:
    agent: Agent
    rel_type: str
    rel_id:str

    def set_arg(self, argn):

        # if the relation starts with the agent pointer,
        # then its type is out
        if argn == "arg1":
            self.out = True
        else:
            self.out = False


class BigBubble:

    def __init__(self, event = None):
        #print("BigBubble")
        self.bubble_ = Bubble(event)
        self.little_bubbles = []
        self.idx = 0 

    def add_bubble(self, event):
        #print("add_bubble")
        self.bubble_ = Bubble(event)

        return self.bubble_
        
    def add_little_bubble(self, event_lst):

        #print("add_little_bubble")
        little_bubble = Bubble(event_lst)
        self.little_bubbles.append(little_bubble)

        return little_bubble

    def sort_by_offset(self):
        # sort little bubbles by offset
        tmp_bubbles = [(b,b.event.offset) for b in self.little_bubbles]
        tmp_bubbles = sorted(tmp_bubbles, key=lambda x:x[1])
        self.little_bubbles = [b for (b, offset) in tmp_bubbles]

    def create_relations(self, map_ann, relations_set, type_rel_lst = []):

        # add big bubble relations
        self.bubble_.add_relations(map_ann,relations_set, type_rel_lst)

        
        # create relations for the little bubbles
        for bubble in self.little_bubbles:
            bubble.add_relations(map_ann, relations_set, type_rel_lst)

class BubbleRelation:
    def __init__(self, edge_type, bubble, argn):
        self.edge_type = edge_type
        self.bubble_pointer = bubble

        # if the relation starts with bubble pointer,
        # then its type is out
        if argn == "arg1":
            self.out = True
        else:
            self.out = False

class BubbleMap:

    def __init__(self):
        self.map = {} # sent_id -> BigBubble


    def collect_bubble_tokens(self,tok_lst, sent_id, map_ann):

        bubble_toks = []

        for tok in tok_lst:

            # only events
            if is_event(tok):

                # if this is a new relation, but with an already existing bubble
                # then we dont need to create a new bubble
                if tok.id_ann[0] in map_ann:
                    continue

                # it is a little bubble of this big bubble, if 
                # it is in the same sentence
                if tok.sent_id == sent_id:
                    bubble_toks.append(tok)

        return bubble_toks

    def build_map(self, tok_lst: [TokenCorpus], type_big_bubble: str, type_rel_lst: [str]) -> None:

        # for each id annotation get the related events
        event_lst = get_all_events(tok_lst)

        # get only eligible big bubble events (with temporal links identity events)
        event_lst, not_big_bubble_events = sieve_bbubbles_events(event_lst, type_big_bubble)

        event_lst = [event_lst[id_ann] for id_ann in event_lst]

        # map the bubble by the id of annotations, 
        # then, after create the bubble, we create the bubble relations
        map_ann = {} 

        # create the big bubbles
        for event in event_lst:

            big_bubble = BigBubble(event)
            sent_id = event[0].sent_id
            id_ann = event[0].id_ann[0]

            if sent_id in self.map:
                new_sent_id = sent_id + 0.1
                while (new_sent_id in self.map):
                    new_sent_id += 0.1

                big_bubble.bubble_.event.sent_id = new_sent_id
                for tok in big_bubble.bubble_.event_chain:
                    tok.sent_id = new_sent_id

                self.map[new_sent_id] = big_bubble
            else:
                self.map[sent_id] = big_bubble

            map_ann[id_ann] = big_bubble
            
        for event in event_lst:
            id_ann = event[0].id_ann[0]
            sent_id = event[0].sent_id    
            
            for r in event[0].relations:
                rel_sent_id = r.toks[0].sent_id
                rel_id_ann = r.toks[0].id_ann[0]

                # if the event related to the current event is in the same sentence,
                # then add it as a little bubble. Otherwise it is a event that is big bubble,
                # which was already created before or will be created as a little bubble of 
                # of another big bubble
                if int(rel_sent_id) == int(sent_id) and is_type_rel(r.rel_type, type_rel_lst):

                    # add little bubbles (only events)
                    bubble_toks = self.collect_bubble_tokens(r.toks, sent_id, map_ann)

                
                    if len(bubble_toks) > 0:
                        bbubble = self.map[sent_id]
                        little_bubble = bbubble.add_little_bubble(bubble_toks)
                        lbubble_id_ann = bubble_toks[0].id_ann[0]
                        
                        map_ann[lbubble_id_ann] = little_bubble
                        #print()


        # events that can be little bubbles, but are not linked to big bubbles 
        # TODO: mas nao deveriam so eventos que estivessem ligados por tlink?
        for id_ann in not_big_bubble_events:


            sent_id = not_big_bubble_events[id_ann][0].sent_id
            # it is an already created bubble AND
            if id_ann in map_ann:
                continue

            #...and there is no big bubble container, ignore this event
            if sent_id not in self.map:
                continue
             
            #print("-->", sent_id, id_ann, not_big_bubble_events[id_ann])

            bbubble = self.map[sent_id]

            little_bubble = bbubble.add_little_bubble(not_big_bubble_events[id_ann])
            map_ann[id_ann] = little_bubble
   


        # keep track of the relations (id) already created
        relations_set = set()
        for sent_id in self.map:
            bbubble = self.map[sent_id]
            bbubble.create_relations(map_ann, relations_set, type_rel_lst)    
            bbubble.sort_by_offset()

        # now, add the relations regarding the agents
        type_agent_rel = ["OLINK_objIdentity","OLINK_partOf","OLINK_memberOf","SRLINK_agent","SRLINK_medium"]
        for sent_id in self.map:
            bbubble = self.map[sent_id]
            bbubble.bubble_.add_agent_relations(map_ann, relations_set, type_agent_rel)

    def to_json(self, output_file_name):

        data = {}

        for sent_id in self.map:
            data[sent_id] = {}
            big_bubble = self.map[sent_id]

            # the info about the big bubble
            data[sent_id]["bubble"] = {}
            data[sent_id]["bubble"]["id"] = big_bubble.bubble_.event.id_ann[0]
            data[sent_id]["bubble"]["name"] = big_bubble.bubble_.name
            data[sent_id]["bubble"]["value"] = big_bubble.bubble_.get_event_text()

            # little bubbles information
            data[sent_id]["bubbles"] = {}

            for lbubble in big_bubble.little_bubbles:
                data[sent_id]["bubbles"]["bubble"] = {}
                data[sent_id]["bubbles"]["bubble"]["id"] = lbubble.event.id_ann[0]
                data[sent_id]["bubbles"]["bubble"]["name"] = lbubble.name
                data[sent_id]["bubbles"]["bubble"]["value"] = lbubble.get_event_text()

                for rel in lbubble.relations:
                    data[sent_id]["bubbles"]["bubble"]["relations"] = {}
                    if isinstance(rel.bubble_pointer, Bubble):
                        data[sent_id]["bubbles"]["bubble"]["relations"]["next"] = rel.bubble_pointer.event.id_ann[0]
                    else:
                        data[sent_id]["bubbles"]["bubble"]["relations"]["next"] = rel.bubble_pointer.bubble_.event.id_ann[0]
                    data[sent_id]["bubbles"]["bubble"]["relations"]["type"] = rel.edge_type
                    data[sent_id]["bubbles"]["bubble"]["relations"]["out"] = rel.out

            # relations
            for rel in big_bubble.bubble_.relations:
                data[sent_id]["relations"] = {}

                if isinstance(rel.bubble_pointer, Bubble):
                    data[sent_id]["relations"]["next"] = rel.bubble_pointer.event.id_ann[0]
                else:
                    data[sent_id]["relations"]["next"] = rel.bubble_pointer.bubble_.event.id_ann[0]

                data[sent_id]["relations"]["type"] = rel.edge_type
                data[sent_id]["relations"]["out"] = rel.out

        with open(output_file_name, "w") as fd:
            json.dump(data, fd, indent=4)






"""
    AllenNLP annotator

    Used for:
        - Coref resolution
            'en' : 'https://storage.googleapis.com/allennlp-public-models/coref-spanbert-large-2020.02.27.tar.gz'
"""

import pandas as pd
import numpy as np
from itertools import zip_longest

import nltk
from nltk.tokenize import sent_tokenize

from text2story.annotators.ALLENNLP import my_model,my_reader
from text2story.core.utils import bsearch_tuplelist

from allennlp.predictors.predictor import Predictor
from allennlp.models.archival import load_archive
from allennlp_models.structured_prediction.predictors import SemanticRoleLabelerPredictor
from allennlp.models import Model
from overrides import overrides
from typing import List, Iterator, Optional
from nltk import tokenize, download
download('punkt')

from allennlp.data import DatasetReader, Instance
from allennlp.data.vocabulary import Vocabulary
from allennlp.common.util import JsonDict

import spacy

import urllib.request
import os
import torch
from shutil import copyfile
import json


### Global Variables ###

SRL_TYPE_MAPPING = {
    "TMP": "time",
    "LOC": "location",
    "ADV": "theme",  # Adverbial
    "MNR": "manner",
    "CAU": "cause",
    "PRP":"cause",
    "EXT": "theme",  # should be attribute -> changed for compatibility
    "DIS": "theme",  # connection of two expressions -> should be theme -> changed for compatibility
    "PNC": "purpose",
    "PR": "purpose",
    "NEG": "theme",  # should be attribute -> changed for compatibility | NEG = negation
    "DIR": "path",  # should be setting -> changed for compatibility
    "MOD": "instrument",  # MOD = Modal
    "PRD": "theme",  # Secondary predicate -> should be attribute -> changed for compatibility
    "ADJ": "theme",
    "COM": "agent",  # COM = Comitative -> used to express accompaniment
    "GOL": "goal",  # GOL = goal
    "REC": "instrument",  # REC = reciprocal
    "PAS": "passive", # passive verb
    "NSE": "agent",
    "TML": "time",
    "ASP": "theme"
}

pipeline = {}

### Portuguese SRL predictor ###

@Predictor.register("predictor_pt")
class predict(SemanticRoleLabelerPredictor):
    def __init__(
        self, model: Model, dataset_reader: DatasetReader
    ) -> None:
        super().__init__(model, dataset_reader, "pt_core_news_lg")

    @overrides
    def load_line(self, line: str) -> Iterator[str]:
        for sentence in tokenize.sent_tokenize(line):
            yield sentence

    @overrides
    def dump_line(self, outputs) -> str:
        output_file=open("output.txt", "a",encoding="utf-8")
        return json.dump(outputs, output_file,ensure_ascii=False)



    @overrides
    def _sentence_to_srl_instances(self, sentence):
        new_sent=""
        for char in sentence:
            new_sent+=char if char!="-" else " -"
        tokens = self._tokenizer.tokenize(new_sent)
        return self.tokens_to_instances(tokens)

    @overrides
    def predict_json(self, inputs: str):
        instances = self._sentence_to_srl_instances(inputs)
        if not instances:
            return inputs.split()
        return self.predict_instances(instances)

    @overrides
    def predict_instances(self, instances: List[Instance]) -> JsonDict:
        outputs = self._model.forward_on_instances(instances)
        results = {"verbs": [], "words": outputs[0]["words"]}
        for output in outputs:
            tags = output["tags"]
            description = self.make_srl_string(output["words"], tags)
            results["verbs"].append(
                {"verb": output["verb"], "description": description, "tags": tags}
            )
        return results

def download_srl_pt_model():
    current_path = os.path.dirname(os.path.abspath(__file__))
    model_name = "srl-pt_bertimbau-base"

    vocab = Vocabulary()
    vocab_dir = os.path.join(current_path, "Models")
    vocab_dir = os.path.join(vocab_dir, model_name)
    vocab_dir = os.path.join(vocab_dir, "Vocabulary")
    vocab = vocab.from_files(vocab_dir)   

    layer_dir = os.path.join(current_path, "Models")
    layer_dir = os.path.join(layer_dir, model_name)
    layer_dir = os.path.join(layer_dir, "linear_layer.pt")
    ll = torch.load(layer_dir)

    model = my_model.SrlMyModel(vocab = vocab, bert_model = os.path.join("liaad", model_name))
    model.tag_projection_layer.load_state_dict(ll)

    model_dir = os.path.join(current_path, "cache")
    if not(os.path.exists(model_dir)):
        os.mkdir(model_dir)

    model_dir = os.path.join(model_dir, model_name)
    if not(os.path.exists(model_dir)):
        os.mkdir(model_dir)

    torch.save(model.state_dict(), os.path.join(model_dir, "weights.th"))
    model.vocab.save_to_files(os.path.join(model_dir, "vocabulary"))

    config_dir = os.path.join(current_path, "Models")
    config_dir = os.path.join(config_dir, model_name)
    config_dir = os.path.join(config_dir, "config.json")
    copyfile(config_dir, os.path.join(model_dir, "config.json"))

### methods to load models and extract narrative elements  ###

def load(lang):

    if lang == "en":
        load_en()
    elif lang == "pt":
        load_pt()


def load_en():

    nltk.download('omw-1.4')

    current_path = os.path.dirname(os.path.abspath(__file__))

    url_coref = 'https://storage.googleapis.com/allennlp-public-models/coref-spanbert-large-2021.03.10.tar.gz'
    coref_path = os.path.join(current_path, "cache")
    if not(os.path.exists(coref_path)):
        os.mkdir(coref_path)

    coref_path = os.path.join(coref_path, "coref-spanbert-large-2021.03.10.tar.gz")

    if not(os.path.exists(coref_path)):
        print("Downloading %s. It can take some time..." % coref_path)
        urllib.request.urlretrieve("https://storage.googleapis.com/allennlp-public-models/coref-spanbert-large-2021.03.10.tar.gz",\
                coref_path)

    pipeline['coref_en'] = Predictor.from_path(coref_path)

    url_srl = 'https://storage.googleapis.com/allennlp-public-models/structured-prediction-srl-bert.2020.12.15.tar.gz'
    srl_path = os.path.join(current_path, "cache")
    srl_path = os.path.join(srl_path, "structured-prediction-srl-bert.2020.12.15.tar.gz")

    if not(os.path.exists(srl_path)):
        print("Downloading %s. It can take some time..." % srl_path)
        urllib.request.urlretrieve("https://storage.googleapis.com/allennlp-public-models/structured-prediction-srl-bert.2020.12.15.tar.gz",\
                srl_path)


    pipeline["srl_en"] = Predictor.from_path(srl_path)
    # regular expression to capture SRL tags
    pipeline["srl_re_tags"] = r".*[ARG][0-9]|ARGM"
    pipeline["ARGM_STR"] = "ARGM"

    if not(spacy.util.is_package('en_core_web_lg')):
        spacy.cli.download('en_core_web_lg')

    pipeline['en'] = spacy.load('en_core_web_lg')

def load_pt():

    current_path = os.path.dirname(os.path.abspath(__file__))

    srl_pt_model_path = os.path.join(current_path, "cache")
    srl_pt_model_path = os.path.join(srl_pt_model_path, "srl-pt_bertimbau-base")

    if not(os.path.exists(srl_pt_model_path)):
        download_srl_pt_model()

    archive = load_archive(srl_pt_model_path)

    pipeline["srl_pt"] = Predictor.from_archive(
            archive, "predictor_pt", dataset_reader_to_load="train"
        )

    # regular expression to capture SRL tags
    pipeline["srl_re_tags"] = r".*[A][0-9]|AM"

    # since allennlp does not have NER, download the spacy model
    if not(spacy.util.is_package('pt_core_news_lg')):
        spacy.cli.download('pt_core_news_lg')
    pipeline["ARGM_STR"] = "B-AM"

    pipeline["pt"] =  spacy.load('pt_core_news_lg')

def _normalize_sent_tags(sentence_df):
    """
    Normalize the frames retrieved from the SRL from one sentence.
    Each word must have only one label.

    @param sentence_df: DataFrame of the SRL each column is a word from the sentence and each row is the results of SRL
    for one frame.
    @return: List of the normalized tags
    @return: List of booleans of whether the tag is the beginning of the argument.
    """
    normalized_tags, begin_tags = [], []
    if sentence_df.size  == 0:
        return normalized_tags, begin_tags

    for col in np.arange(len(sentence_df.columns)):
        word_vals = sentence_df.iloc[:, col]

        word_vals = word_vals[word_vals != "O"]
        if word_vals.shape[0] == 1:
            normalized_tags.append(word_vals.iloc[0])
            begin_tags.append(word_vals.iloc[0].startswith("B"))
            continue
        verb_words = word_vals[word_vals.isin(["I-V", "B-V"])]
        if verb_words.shape[0] != 0:  # a) - verbo
            normalized_tags.append(verb_words.iloc[0])
            begin_tags.append(False)  # Event
            continue
        # b) - ARGM e ARG (o último e mais especifico tem prio)
        # TODO: em português as tags estao ligeiramente diferentes: 
        # B-AM-TMP, B-A1, I-A2, etc
        arg_words = word_vals[word_vals.str.contains(pipeline["srl_re_tags"])]
        if arg_words.shape[0] != 0:
            normalized_tags.append(arg_words.iloc[-1])  # desempate entre dois ARGM-X diferentes
            begin_tags.append(arg_words.iloc[-1].startswith("B"))
            continue
        else:
            print("\nNORMALIZATION ERROR - MULTIPLE TAG VALUES FOUND FOR WORD.")
            print(word_vals.values) 
            print(sentence_df)
    #print(">>>",begin_tags, sentence_df)
    #print("-->",normalized_tags)
    return normalized_tags, begin_tags


def _find_events(normalized_tags, verb_tags, event_threshold=2):
    """
    Find words that belong to the same event.
    Each event can have a number of arguments between verbs and still be considered the same event.

    @param normalized_tags: result of normalized_sent_tags - a list of SRL tags for each word
    @param verb_tags: list of SRL tags for the algorithm to classify as event tags
    @param event_threshold: Threshold of non-verb arguments that can be found between verbs
    and still be considered part of the same event. n_args = event_threshold - 1
    @return: Boolean list of whether or not the word is part of an event
    """
    event_tags = []
    event_continue, event_begin = False, False
    for i, tag in zip_longest(np.arange(len(normalized_tags) - event_threshold), normalized_tags):
        if i is not None:
            if ("ARGM" in tag) & (normalized_tags[i + 1] in verb_tags):
                event_tags.append(True)
                event_begin = True
                continue

        if event_continue:
            event_tags.append(True)
        elif tag in verb_tags:
            event_tags.append(True)
            event_begin = True
        else:
            event_tags.append(False)

        if i is not None:
            conds = []
            for j in np.arange(1, event_threshold + 1):
                conds.append(normalized_tags[i + j] in verb_tags)

            if event_begin & any(conds):
                event_continue = True
            else:
                event_continue = False
                event_begin = False
        else:
            event_continue = False
            event_begin = False

    return event_tags


def _find_actors(begin_tags, event_tags):
    """
    Finds words that belong to the same actor or event and categorize them into actors.
    Each different argument represents a different actor. Arguments start at the begin tag (B).

    @param begin_tags: result of normalized_sent_tags - boolean list of tags that begin SRL arguments
    @param event_tags: result of find_events - boolean list of words that represent events
    @return: List of the actors that are represented by each word in the sentence
    """
    actor, i, actor_tags, event, is_event = False, 0, [], 1, False
    for btag, etag in zip(begin_tags, event_tags):
        if etag:
            is_event = True
        elif (not etag) & is_event:
            is_event = False
            event += 1

        if btag & (not etag):
            actor = True
            i += 1
            actor_tags.append("T" + str(i))
            continue

        if etag:
            actor_tags.append("EVENT" + str(event))
            actor = False
            continue

        if (not btag) & actor:
            actor_tags.append("T" + str(i))
        else:
            if len(actor_tags) > 0:
                actor_tags.append(actor_tags[-1])  # In case everything else fails, just join with previous actor
            else:
                # TODO: when actor_tags is empty what to do?
                i += 1
                actor_tags.append("T" + str(i))

    return actor_tags


def _srl_by_actor(srl_by_token, text, char_offset):
    """
    Organizes actors as words or expressions in the full text with their respective semantic role and character span.

    The semantic role is "EVENT" for event arguments and the SRL result for other actors.
    If the SRL result is a modifier, the most common in the actor is taken into account.

    The character span is a tuple - (start_char, end_char) -
    where the "start_char" is where the actor starts in the text and the "end_char" is where the actor ends.

    @param srl_by_token: DataFrame containing the results for the rest of the pipeline - namely, the actor references
    @param text: The full text to be annotated
    @param char_offset: The current char position in the text. Each word in the text increments it
    @return: Dict list with the actor, its semantic role and the its character position span in the text

    @note: CAN MALFUNCTION IF SRL DOES NOT FIND FRAMES IN EVERY SENTENCE.
    The char positions are checked in the entire text, if there is a word before in a sentence not recognized by the SRL
    it will malfunction slightly (give wrong position values).
    """
    result_list = []
    ARGM_STR = pipeline["ARGM_STR"]
    for actor in srl_by_token["actor"].unique():
        rows = srl_by_token[srl_by_token["actor"] == actor]
        tags = rows["tag"]
        if actor.startswith("EVENT"):
            sem_role_type = "EVENT"
        elif any(ARGM_STR in tag for tag in tags):
            argm_tags = [tag for tag in tags if ARGM_STR in tag]
            #print("-->",argm_tags[0].split("-")[-1])
            sem_role_type = SRL_TYPE_MAPPING[argm_tags[0].split("-")[-1]]
        else:
            sem_role_type = "PARTICIPANT"

        char_spans = []
        for token in rows.index:
            char_offset = text.find(token, char_offset)
            char_spans.append(char_offset)
            char_offset += len(token)

        result_list.append({
            "actor": ' '.join(rows.index), "sem_role_type": sem_role_type,
            "char_span": (char_spans[0], char_spans[-1] + len(token))
        })

    return result_list, char_offset


def _make_srl_df(lang, text):
    """
    Make a pandas DataFrame with the results from the SRL for each sentence in the text.
    Each row of the DataFrame is a frame from the SRL.

    @param text: The full text to annotate
    @return: List of pandas DataFrames with the contents of the SRL (values) for each frame (row) by word (column)
    """
    sentences = sent_tokenize(text)

    srl = []
    # se for um texto muito grande, da problema. O Allennlp tem um limite aqui
    for sent in sentences:

        toks = nltk.word_tokenize(sent)

        # allennlp has a limitation in the srl pipeline, they only accept sentences
        # with at most 512 tokens. So in here there is a test, and if necessary it splits
        # the sentence in batches of tokens
        if len(toks) > 256:

            result = []
            batch = 1
            len_batch = 256
            start_batch = 0
            end_batch = len_batch * batch

            while end_batch < len(toks):

                batch_sent = " ".join(toks[start_batch:end_batch])
                if abs(end_batch - start_batch) < 5: 
                    batch = batch + 1
                    start_batch = end_batch
                    end_batch = len_batch * batch
                    continue


                if lang == "en":
                    result = pipeline["srl_en"].predict(sentence=batch_sent)
                elif lang == "pt":
                    try:
                         result = pipeline["srl_pt"].predict_json(batch_sent)
                    except RuntimeError:
                        print("Warning: Runtime Error -- %s" % batch_sent)
                        break
                    # there is a bug for some sentences, that there is no result from the srl
                    if isinstance(result, list):
                        continue

                batch = batch + 1
                start_batch = end_batch
                end_batch = len_batch * batch
                srl.append(result)
        else:
            if lang == "en":
                result = pipeline["srl_en"].predict(sentence=sent)
            elif lang == "pt":
                result = pipeline["srl_pt"].predict_json(sent)
                # there is a bug for some sentences, that there is no result from the srl
                if isinstance(result, list):
                    continue
            srl.append(result)

    dfs_by_sent = []
    for sentence in srl:
        tags_by_frame = pd.DataFrame(columns=sentence["words"])
        for i, frame in zip(np.arange(len(sentence["verbs"])), sentence["verbs"]):
            tags_by_frame.loc[i] = frame["tags"]

        if tags_by_frame.shape[0] != 0:
            dfs_by_sent.append(tags_by_frame)

    return dfs_by_sent


def _srl_pipeline(sentence_df, text, char_offset, verb_tags, event_threshold=3):
    """
    Pipeline to retrieve actors and events by their order in the text, with their semantic roles and character spans.
    Each iteration of the pipeline takes a dataframe of the SRL results for a sentence -> as returned by _make_srl_df.

    @param sentence_df: The DataFrame with the SRL results with the structure shown in _make_srl_df
    @param text: The full text to be annotated
    @param char_offset: The current character position offset to start looking for character spans (recursive)
    @param verb_tags: SRL tags that represent verbs -> defaults to ["B-V", "I-V"]
    @param event_threshold: Threshold of non-verb arguments that can be found between verbs
    and still be considered part of the same event. number_of_args = event_threshold - 1

    @return: df_by_actor -> Pandas DataFrame with each actor, their semantic roles and character spans
    character_offset -> The current character position offset in the full text
    """
    # 2. Remove words out of vocabulary in every frame #
    oov_mask = []
    for name, values in sentence_df.items():
        oov_mask.append(any(~(values == "O")))
    sentence_df = sentence_df.loc[:, oov_mask]
    sentence_df = sentence_df.apply(lambda x: x.sort_values(ascending=False).values)  # Begin tags in the end always

    # 3. Normalize SRL tags for each word in the sentence dataframe #
    normalized_tags, begin_tags = _normalize_sent_tags(sentence_df)

    # 4. Define events #
    event_tags = _find_events(normalized_tags, verb_tags=verb_tags, event_threshold=event_threshold)

    #print("---", event_tags)
    #print()
    # 5. Find and categorize expressions into actors #
    actor_tags = _find_actors(begin_tags, event_tags)

    # Putting together a dataframe for the characteristics of each word in this sentence
    if sentence_df.size == 0:
        return pd.DataFrame(), 0
    df = pd.DataFrame({
        "tag": normalized_tags, "is_begin_tag": begin_tags, "is_event": event_tags, "actor": actor_tags
    }, index=sentence_df.columns)

    # 6. Find semantic roles and character spans for each actor in the sentence #
    df_by_actor, character_offset = _srl_by_actor(df, text, char_offset)

    return df_by_actor, character_offset

def get_participant_tags(text, participant_lst, lang):
    """
    it adds POS tags and NER tags to the participant list

    @param text: the original text to be processed as narrative
    @param list[tuple[int,int]]: A list of character span of the participant
    candidates
    Returns
    ------
    list[tuple[tuple[int, int], str, str]]
       a list of participants of narratives, where the first is the character
       span of that participant, the second elements is the POS tag, anf third
       element is the NER tag, if it is a entity
    """
    participant_lst = sorted(participant_lst, key=lambda x:x[0])
    tagged_participants = {}

    doc = pipeline[lang](text)

    for tok in doc:
        pos = bsearch_tuplelist(tok.idx, participant_lst)
        if pos != -1:
            # only update the tag value with the lexical head of the
            # participant
            if pos not in tagged_participants:
                offset_participant = participant_lst[pos]
                if tok.ent_type_ == '':
                    tagged_participants[pos] = (offset_participant, tok.pos_, 'UNDEF')
                else:
                    tagged_participants[pos] = (offset_participant, tok.pos_,tok.ent_type_.lower().capitalize())

    tagged_participants_lst = []
    # it is possible that some participant was not found by the bsearch?
    for idx in range(len(participant_lst)):
        if idx not in tagged_participants:
            tagged_participants_lst.append((participant_lst[idx],"UNDEF","UNDEF"))
        else:
            tagged_participants_lst.append(tagged_participants[idx])

    return tagged_participants_lst

def extract_actors(lang, text):
    """
    Main function that applies the SRL pipeline to extract participant entities from each sentence.

    @param lang: The language of the text
    @param text: The full text to be annotated
    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        the list of actors identified where each actor is represented by a tuple
    """
    # 1. DATAFRAME WITH THE SRL RESULTS OF EVERY FRAME FOR EACH SENTENCE IN THE TEXT #
    dfs_by_sent = _make_srl_df(lang, text)

    # FIND EVENTS - PIPELINE #
    character_offset, srl_actors_list = 0, []
    for sent_df in dfs_by_sent:
        df_by_actor, character_offset = _srl_pipeline(sent_df, text, character_offset, verb_tags=["B-V", "I-V"],
                                                      event_threshold=2)

        srl_actors_list.append(df_by_actor)


    srl_actors_list = [item for sublist in srl_actors_list for item in sublist]  # flatten list
    srl_df = pd.DataFrame(srl_actors_list)
    if len(srl_df) == 0:
        return []

    participants_df = srl_df[srl_df["sem_role_type"] == "PARTICIPANT"]
    participants_lst = participants_df["char_span"].values.tolist()

    # TODO: these participants still dont have POS and NER tags
    participants_tagged_lst = get_participant_tags(text, participants_lst, lang)
    #participants_lst = [ (p, "NOUN", "Per") for p in participants_lst]

    return participants_tagged_lst


def extract_events(lang, text):
    """
    Main function that applies the SRL pipeline to extract event entities from each sentence.
    Joins every event actor from each sentence in the text.

    @param lang: The language of the text
    @param text: The full text to be annotated

    @return: Pandas DataFrame with every event entity and their character span
    """
    # 1. DATAFRAME WITH THE SRL RESULTS OF EVERY FRAME FOR EACH SENTENCE IN THE TEXT #
    dfs_by_sent = _make_srl_df(lang, text)

    # FIND EVENTS - PIPELINE #
    character_offset, srl_actors_list = 0, []
    for sent_df in dfs_by_sent:
        df_by_actor, character_offset = _srl_pipeline(sent_df, text, character_offset, verb_tags=["B-V", "I-V"],
                                                      event_threshold=3)
        srl_actors_list.append(df_by_actor)

    srl_actors_list = [item for sublist in srl_actors_list for item in sublist]  # flatten list
    srl_df = pd.DataFrame(srl_actors_list)
    if len(srl_df) == 0:
        return []
    return srl_df[srl_df["sem_role_type"] == "EVENT"]


def extract_semantic_role_links(lang, text):
    """
    Main function that applies the SRL pipeline to extract the semantic role links between actors and events.
    Joins the Semantic Role Links from each sentence in the text.

    @param lang: The language of the text
    @param text: The full text to be annotated

    @return: List of pandas DataFrames that contains the SRL for each actor in each sentence.
    """
    dfs_by_sent = _make_srl_df(lang, text)

    character_offset, srl_by_sentence = 0, []
    for sent_df in dfs_by_sent:
        df_by_actor, character_offset = _srl_pipeline(sent_df, text, character_offset, verb_tags=["B-V", "I-V"],
                                                      event_threshold=3)
        srl_by_sentence.append(pd.DataFrame(df_by_actor))

    return srl_by_sentence



def extract_objectal_links(lang, text):
    prediction = pipeline['coref_en'].predict(document=text)

    cluster_indexes_list = prediction["clusters"] # Indexes are token spans, we need character spans
   # Compute the character spans
    character_offset = 0
    character_span = []
    for token in prediction["document"]:
        character_offset = text.find(token, character_offset)
        character_span.append((character_offset, character_offset + len(token)))
        character_offset += len(token)

   # Convert the clusters to character spans
    for cluster in cluster_indexes_list:
        for i in range(len(cluster)):
            start_token_span = cluster[i][0]
            end_token_span = cluster[i][1]

            char_offset = (character_span[start_token_span][0], character_span[end_token_span][1])

            cluster[i] = char_offset

    return cluster_indexes_list



"""
    text2story/core/utils module

    Functions
    ---------
    chunknize_actors(annotations)
        converts the result from the 'extract_actors' method, implemented by the annotators supported, to a list the chunknized list of actors
        to do this conversion, the IOB in the NE tag is used
"""

from itertools import tee

import numpy as np
from typing import Tuple, List


def bsearch_tuplelist(x: int, xs: List[Tuple[int, int]]) -> int:
    """
    Use binary search to check if x is whithin some interval of a tuple list
    xs.

    if x is not in none of intervals, it return -1, otherwise returns de
    position of the interval.
    @type xs: object

    """
    b = 0
    e = len(xs) - 1
    m = (b + e) // 2
    pos = -1

    while b <= e:
        (x_b, x_e) = xs[m]

        if x>= x_b and x<= x_e:
            return m
        elif x < x_b:
            e = m - 1
        elif x > x_e:
            b = m + 1

        m = (b + e) // 2

    return pos



def join_tokens(tok_lst):
    pronomes = ["me", "te", "lhe", "o", "a", "la", "lo", "lho", "lha", "nos"]
    parentheses = ["(", "{"]
    punctuation = [",", ".", ":", ";", "?", "!", ")", "}"]

    # first, find all hyphen in the token list
    array_tok_lst = np.array(tok_lst)
    indices = np.where(array_tok_lst == "-")[0]
    indices = list(indices)

    if indices != []:
        tok_join = ""
        if len(tok_lst) > 0:
            i = 0
            while i < len(tok_lst):
                if i in indices:
                    if i + 1 < len(tok_lst) and \
                            tok_lst[i + 1].lower() in pronomes:
                        tok_join = tok_join.strip() + tok_lst[i]
                        i = i + 1

                if tok_lst[i] in parentheses:
                    tok_join = tok_join + tok_lst[i]
                else:
                    if tok_lst[i] in punctuation:
                        tok_join = tok_join.rstrip() + tok_lst[i] + " "
                    else:
                        tok_join = tok_join + tok_lst[i] + " "
                i = i + 1

        return tok_join.strip()
    else:
        tok_join = ""
        if len(tok_lst) > 0:
            i = 0
            while i < len(tok_lst):

                if tok_lst[i] in parentheses:
                    tok_join = tok_join + tok_lst[i]
                else:
                    if tok_lst[i] in punctuation:
                        tok_join = tok_join.rstrip() + tok_lst[i] + " "
                    else:
                        tok_join = tok_join + tok_lst[i] + " "
                i = i + 1

        return tok_join.strip()


def chunknize_actors(annotations):
    """
    Parameters
    ----------
    annotations : list[tuple[tuple[int, int], str, str]]
        list of annotations made by some tool for each token

    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        the list of actors identified where each actor is represented by a tuple
    """

    actors = []

    ready_to_add = False

    prev_ne_tag = ''

    for ann in annotations:
        token_character_span, token_pos_tag, token_ne_iob_tag = ann

        # token_ne_iob_tag = 'I-PER', then token_ne_iob_tag[2:] == 'PER'
        if token_ne_iob_tag.startswith("B") or (
                token_ne_iob_tag.startswith("I") and token_ne_iob_tag[2:] != prev_ne_tag):
            # Case we start a new chunk, after finishing another, for instance: Case B-Per, I-Per, B(or I)-Org, then we add the finished actor
            if ready_to_add:
                actor = ((actor_start_offset, actor_end_offset), actor_lexical_head, actor_actor_type)
                actors.append(actor)

            ready_to_add = True
            actor_start_offset = token_character_span[0]
            actor_end_offset = token_character_span[1]
            actor_lexical_head = token_pos_tag if token_pos_tag in ['Noun', 'Pronoun'] else 'UNDEF'
            actor_actor_type = token_ne_iob_tag[2:]

        elif token_ne_iob_tag.startswith("I"):
            # actor_start_offset it's always the same, since it's defined by the first token of the actor
            actor_end_offset = token_character_span[1]
            actor_lexical_head = actor_lexical_head if actor_lexical_head != 'UNDEF' else token_pos_tag if token_pos_tag in [
                'Noun', 'Pronoun'] else 'UNDEF'
            # actor_actor_type it's the same for all tokens that constitute the actor and it's already defined by the first token of the actor

        elif token_ne_iob_tag.startswith("O") and ready_to_add:
            actor = ((actor_start_offset, actor_end_offset), actor_lexical_head, actor_actor_type)
            actors.append(actor)
            ready_to_add = False
            # No need to reset the variables, since the first update writes over

        prev_ne_tag = token_ne_iob_tag[2:]

    if ready_to_add:  # If the last token still makes part of the actor
        actor = ((actor_start_offset, actor_end_offset), actor_lexical_head, actor_actor_type)
        actors.append(actor)

    return actors


def pairwise(iterable):
    """
    Iterate through some iterable with a lookahead.
    From the itertools docs recipes - https://docs.python.org/3/library/itertools.html
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)



import re

from text2story.readers.token_corpus import TokenCorpus
from text2story.core.utils import join_tokens

class NodeEvent:

    def __init__(self, event_txt, event_class, event_idx, event_arg, event_sent, id_ann = None):
        self.txt = event_txt
        self.class_type = event_class
        self.idx = event_idx 
        self.argn = event_arg
        self.sent_id = event_sent
        self.id_ann = id_ann

class EdgeEvent:

	def __init__(self, edge_type):
		self.edge_type = edge_type
		self.nodes = []

	def add_node(self, event_txt, event_class, event_idx, event_arg, event_sent, id_ann = None):
		self.nodes.append(NodeEvent(event_txt, event_class, event_idx, event_arg, event_sent, id_ann))

def sieve_bbubbles_events(event_lst, type_big_bubble):
    """
    A sieve for events that has temporal links of identity types
    """

    ans = {}
    not_ans = {}

    for id_ann in event_lst:
        
        for tok in event_lst[id_ann]:


            for attr_item in tok.attr:
 
                ann_type = attr_item[0]
                attr_map = attr_item[1]

                if "Class" not in attr_map:
                    continue

                if attr_map["Class"] == type_big_bubble:
                    is_big_bubble = False

                    for r in tok.relations:
                        #print(r.rel_type, id_ann, tok.text)
                        if is_event(r.toks[0]):
                            if r.rel_type == "TLINK_identity" and\
                                id_ann not in ans and type_big_bubble in get_event_type(r.toks[0]):

                                #toks_ids = [rel_toks.id_ann for rel_toks in r.toks]
                                #print("id_ann: ",id_ann, toks_ids)
                                ans[id_ann] = event_lst[id_ann]
                                is_big_bubble = True
                                break

                    if not(is_big_bubble):
                        if id_ann in not_ans:
                            not_ans[id_ann].append(tok)
                        else: 
                            not_ans[id_ann] = [tok]
                    
                else:
                    if id_ann in not_ans:
                        not_ans[id_ann].append(tok)
                    else: 
                        not_ans[id_ann] = [tok]

    return ans, not_ans


def get_reporting_events(tok_lst):

    event_lst = {}


    for i in range(len(tok_lst)):
        tok = tok_lst[i]


        for attr_item in tok.attr:

            ann_type = attr_item[0]
            attr_map = attr_item[1]

            if "Class" in attr_map and attr_map["Class"] == "Reporting": # it is an event token

                for id_ann in tok.id_ann:

                    if id_ann in event_lst:
                        event_lst[id_ann].append(tok)
                    else:
                        event_lst[id_ann] = [tok]
                break
    return event_lst


def get_all_events(tok_lst):

    event_lst = {}


    for i in range(len(tok_lst)):
        tok = tok_lst[i]


        for attr_item in tok.attr:

            ann_type = attr_item[0]
            attr_map = attr_item[1]

            if "Class" in attr_map: # it is an event token

                for id_ann in tok.id_ann:

                    if id_ann in event_lst:
                        event_lst[id_ann].append(tok)
                    else:
                        event_lst[id_ann] = [tok]
                break
            	

    return event_lst

def get_txt_events(event_lst):

    event_txt_lst = []

    for id_ann in event_lst:
        text_tokens = [e.text for e in event_lst[id_ann]]
        event_txt = " ".join(text_tokens)
        event_txt_lst.append(event_txt)

    return event_txt_lst

        
def get_hist_report_events(tok_lst):

    event_lst = get_all_events(tok_lst)
    hist_reporting = {}
    for e in event_lst:

        rel_type = get_event_type(e)
        if "Reporting" in rel_type:
            tok_event = e.text.lower()
            if tok_event in hist_reporting:
                hist_reporting[tok_event] += 1
            else:
                hist_reporting[tok_event] = 1
    return hist_reporting

def map_events_bysent(tok_lst):

    event_lst = get_all_events(tok_lst)
    map_sent = {}
    for e in event_lst:

        primary_type = get_event_type(e)[0]

        if e.sent_id in map_sent:
            if primary_type in map_sent[e.sent_id]:
                map_sent[e.sent_id][primary_type] += 1
            else:
                map_sent[e.sent_id][primary_type] = 1
        else:
            map_sent[e.sent_id] = {primary_type:1}

    return map_sent

def is_event(event):

    for attr_item in event.attr:

        ann_type = attr_item[0]
        attr_map = attr_item[1]

        if ann_type.lower() == "event":
            return True
    return False


def get_event_type(event):

    event_type_lst = []

    
    for attr_item in event.attr:

        ann_type = attr_item[0]
        attr_map = attr_item[1]

        if "Class" in attr_map: 
            event_type_lst.append(attr_map["Class"])
                
    #print("-->", event_type_lst)
    return event_type_lst



"""
	text2story.core.entity_structures

	Entity structures classes (Actor, TimeX and Event)
"""


class ActorEntity:
    """
    Representation of an actor entity.

    Attributes
    ----------
    text: str
        The textual representation of the actor.
    character_span: tuple[int, int]
        The character span of the actor.
    lexical_head: str
        The lexical head of the actor.
        Possible values are: 'Noun' or 'Pronoun'.
    type: str
        The type of the actor.
        Possible values are: 'Per', 'Org', 'Loc', 'Obj', 'Nat' or 'Other'.
    individuation: str
        Stipulation of whether the actor is a set, a single individual, or a mass quantity.
        Possible values are: 'Set', 'Individual' or 'Mass'.
        NOTE: For now, using the label 'Individual' to all actors.
    involvement: str
        The specification of how many entities or how much of the domain are/is participating in an event.
        Possible values are: '0', '1', '>1', 'All' or 'Und'.
        NOTE: For now, using the label '1' to all actors.
    """

    def __init__(self, text, character_span, lexical_head, actor_type):
        self.text = text
        self.character_span = character_span
        self.lexical_head = lexical_head
        self.type = actor_type
        self.individuation = 'Individual'
        self.involvement = '1'


class TimeEntity:
    """
    Representation of a time entity.

    Attributes
    ----------
    text: str
        The textual representation of the time.
    character_span: tuple[int, int]
        The character span of the time.
    value: str
        The value of the time.
        Possible values are: value ::=  Duration | Date | Time | WeekDate | WeekTime | Season | PartOfYear | PaPrFu(Past Present Future Reference)
    type: str
        The type of the time.
        Possible values are: 'Date', 'Time', 'Duration' and 'Set'.
    temporal_function: str
        Possible values are: 'None' or 'Publication_Time'.
    """

    def __init__(self, text, character_span, value, timex_type, temporal_function='Publication_Time'):
        self.text = text
        self.character_span = character_span
        self.value = value
        self.type = timex_type
        self.temporal_function = temporal_function


class EventEntity:
    """
    Representation of an event entity.
        TODO: Annotations (A)
    """

    def __init__(self, text, character_span, event_class="Occurrence", polarity="Pos"):
        self.text = text
        self.character_span = character_span
        self.event_class = event_class
        self.polarity = polarity
        self.factuality = "Factual"
        self.tense = "Pres"



"""
	text2story.core.exceptions

	All defined exceptions raised by the package.
"""

class InvalidLanguage(Exception):
	"""
	Raised if the user specified an invalid/unsupported language.
	"""
	def __init__(self, lang):
		description = ('Invalid language :' + lang)
		
		super().__init__(description)

class InvalidTool(Exception):
	"""
	Raised if the user specified an invalid/unsupported tool.
	"""

	def __init__(self, tool):
		description = ('Invalid tool: ' + tool)
		
		super().__init__(description)



# Version of the text2story package
__version__ = "1.0.0"

from text2story.annotators import load
from text2story.annotators import get_available_tools
from text2story.annotators import get_participant_tools
from text2story.annotators import get_event_tools
from text2story.annotators import get_time_tools
from text2story.annotators import get_srlink_tools

def start(lang):
    load(lang)

def get_tools_name():
    return get_available_tools()

def participant_tool_names():
    return get_participant_tools()

def event_tool_names():
    return get_event_tools()

def time_tool_names():
    return get_time_tools()

def srlink_tool_names():
    return get_srlink_tools()


# Export to out of the package
from text2story.core.narrative import Narrative




"""
	text2story.core.narrative

	Narrative class
"""

from text2story.core.annotator import Annotator
from text2story.core.entity_structures import *
from text2story.core.link_structures import *
from text2story.core.utils import pairwise


class Narrative:
    """
    Representation of a narrative.

    Attributes
    ----------
    lang: str
            the language of the text; supported languages are portuguese ('pt') and english ('en')
    text: str
            the text itself
    publication_time : str
                    the publication time ('XXXX-XX-XX')
    actors: dict{str -> Actor}
            the actors identified in the text.
            each key in the dict, of the form 'T' concatenated with some int, has an actor as a value.
    times: dict{str -> Time}
            the temporal expressions identified in the text.
            each key in the dict, of the form 'T' concatenated with some int, has an time as a value.
    obj_links: dict{str -> ObjectalLink}
            the corefs identified in the text
            each key in the dict, of the form 'R' concatenated with some int, has an coref as a value.

    Methods
    -------
    extract_actors(*tools)
            extracts all the actors in the text using the annotators defined in 'tools', updating self.actors
    extract_timexs(*tools)
            extracts all the timexs in the text using the annotators defined in 'tools', updating self.timexs
    extract_corefs(*tools)
            coreference resolution in the text using the tools 'tools', updating self.obj_rels
            typically, this call increases self.actors since news entities can be identified
    _get_actor_key(char_offset)
            returns the key of the actor with the corresponding character offset or None if such actor wasn't identified before
    _add_actor(char_offset)
            update self.actors by adding the new actor with character offset 'char_offset' and returns the key given to the new actor
    ISO_annotation(file_name)
            outputs ISO annotation in .ann format (txt)
    """

    def __init__(self, lang, text, publication_time):
        """
        Parameters
        ----------
        lang : str
                the language of the text
        text : str
                the text ifself
        publication_time : str
                the publication time ('XXXX-XX-XX')
        """

        self.lang = lang
        self.text = text
        self.publication_time = publication_time

        # Counter to generate a unique ID for every participant
        # TODO: Fix the counter, when repeting some extraction: The counter just keep going up.
        self._id = 1
        self._event_id = 1
        self._rel_id = 1

        self.actors = {}
        self.times = {}
        self.events = {}
        self.obj_links = {}
        self.sem_links = {}

    def extract_actors(self, *tools):
        """
        Parameters
        ----------
        tools : str, ...
                the tools to be used in the annotation

        Returns
        -------
                self.actors updated
        """

        actors = Annotator(tools).extract_actors(self.lang,
                                                 self.text)  # annotations :: [(EntityStartOffset, EntityEndOffset, EntityPOSTag, EntityType)]

        for actor in actors:
            self.actors['T' + str(self._id)] = ActorEntity(self.text[actor[0][0]:actor[0][1]], actor[0], actor[1],
                                                           actor[2])
            self._id += 1

        return self.actors

    def extract_times(self, *tools):
        """
        Parameters
        ----------
        tools : str, ...
                the tools to be used in the annotation

        Returns
        -------
                self.times updated
        """
        times = Annotator(tools).extract_times(
            self.lang, self.text, self.publication_time)  # annotations :: [(TimeStartOffset, TimeEndOffset, TimeType, TimeValue)]

        for time in times:
            self.times['T' + str(self._id)] = TimeEntity(self.text[time[0][0]:time[0][1]], time[0], time[1],
                                                         time[2])
            self._id += 1

        return self.times

    def extract_events(self, *tools):
        """
        Event extraction function to combine different tools of event extraction.
        Currently there is only one tool (AllenNLP) so it just uses that one.

        @param tools: Iterable with the tools to use

        @return: Returns a list of events extracted from the text in the form of EventEntity objects
        """
        events = Annotator(tools).extract_events(self.lang, self.text)
        if len(events) == 0:
            self.events = []
            return []

        for event in events.itertuples():
            self.events["T" + str(self._id)
                        ] = EventEntity(event.actor, event.char_span)
            self._id += 1

        return self.events

    def extract_objectal_links(self, *tools):
        """
        Parameters
        ----------
        tools : str, ...
                the tools to be used in the annotation

        Returns
        -------
                self.obj_rels updated
        """

        # annotations ::
        clusters = Annotator(tools).extract_objectal_links(
            self.lang, self.text)

        for cluster in clusters:
            for i in range(0, len(cluster) - 1):
                e1 = cluster[i]
                e2 = cluster[i + 1]

                # Get the actors
                arg1, arg2 = self._get_actor_key(e1), self._get_actor_key(e2)

                # If one of them wasn't identified in the actor extration, add it as a new actor
                if arg1 == None:
                    arg1 = self._add_actor(e1)
                if arg2 == None:
                    arg2 = self._add_actor(e2)

                # (Type (sameHead, partOf, ...), Arg1, Arg2)
                self.obj_links['R' + str(self._rel_id)
                               ] = ObjectalLink(arg1, arg2)
                self._rel_id += 1

        return self.obj_links

    def extract_semantic_role_links(self, *tools):
        """
        Find semantic role links between extracted actors and events.
        Since the SRL model is different from the NER model, this function maps actors found by the SRL model into
        actors that were already extracted. If the actor was not yet extracted, it adds a new one.

        Links actors to events by text order. If we have ACTOR1 -> EVENT1 -> ACTOR2 in the text,
        then we make two semantic role links - EVENT1 -> ROLE -> ACTOR1 | EVENT1 -> ROLE -> ACTOR2

        @param tools: Iterable of tools to be used

        @return: A dict with the SRL entities by key -> R10: SemanticRoleLink<10>
        """
        srl_by_sentence = Annotator(
            tools).extract_semantic_role_links(self.lang, self.text)

        # FIND OUT IF ARGUMENT OF SRL HAS AN ACTOR RETRIEVED BY THE NER COMPONENT
        # IF NOT, ADD A NEW ACTOR CORRESPONDING TO THE ARGUMENT
        for sentence_df in srl_by_sentence:

            key_list = []
            for row in sentence_df.itertuples():
                event_key = None
                if row.sem_role_type == "EVENT":
                    event_key = self._get_event_key(row.char_span)
                    if event_key is not None:
                        key_list.append(event_key)
                    else:
                        event_key = self._get_event_key(
                            row.char_span, match_type="partial")
                        if event_key is not None:
                            key_list.append(event_key)
                        else:
                            event_key = self._add_event(row.char_span)
                            key_list.append(event_key)
                    continue
                actor_key = self._get_actor_key(
                    row.char_span, match_type="partial")

                if actor_key is not None:
                    key_list.append(actor_key)
                else:
                    if actor_key is None and event_key is None:
                        key_list.append("None")
                #else:
                #    actor_key = self._add_actor(
                #        row.char_span, lexical_head="Noun", actor_type="Other")
                #    key_list.append(actor_key)
            
            sentence_df["key"] = key_list

        # MAKE SEMANTIC ROLE LINK ENTITIES
        for sentence_df in srl_by_sentence:
            for row1, row2 in pairwise(sentence_df.itertuples()):
                sem1 = row1.sem_role_type
                sem2 = row2.sem_role_type

                if (sem1 == "EVENT") | (sem2 == "EVENT"):
                    if sem1 != "EVENT":
                        sem_role, actor, event = sem1, row1.key, row2.key
                    else:
                        sem_role, actor, event = sem2, row2.key, row1.key

                    if actor == "None" or event == "None":
                        continue

                    self.sem_links["R" + str(self._rel_id)] = SemanticRoleLink(
                        actor, event, sem_role.lower())
                    self._rel_id += 1

        return self.sem_links

    def _add_srlink(self, actor, event, sem_role):
        self.sem_links["R" + str(self._rel_id)
                       ] = SemanticRoleLink(actor, event, sem_role.lower())
        self._rel_id += 1

    def _get_actor_key(self, char_span, match_type="exact"):
        """
        Parameters
        ----------
        char_span : (int, int)
                the actor character offset of the actor to find the key
        match_type: str
                type of match for the character span.
                If exact, the span must be exactly the same to match an actor.
                If partial, the span must be partially contained in the actor span to match it.

        Returns
        -------
                the key of the actor with the corresponding character offset
                or None if it doesn't exist
        """
        if match_type == "exact":
            for key in self.actors.keys():
                if self.actors[key].character_span == char_span:
                    return str(key)
        elif match_type == "partial":
            for key in self.actors.keys():
                aSpan = self.actors[key].character_span
                if aSpan[0] <= char_span[0] <= aSpan[1]:
                    return str(key)
                elif aSpan[0] <= char_span[1] <= aSpan[1]:
                    return str(key)
        else:
            raise ValueError(
                f"Parameter match_type must be one of [exact, partial].\nInstead it was {match_type}")

        return None

    def _add_actor(self, char_span, lexical_head="Pronoun", actor_type="Other"):
        """
        Parameters
        ----------
        char_span : (int, int)
                the actor character offset
        lexical_head: str
                The lexical head of the actor: "Noun" or "Pronoun" -> Defaults to "Pronoun"
        actor_type: str
                The type of the actor as expressed by the NER models -> Defaults to "Other"

        Returns
        -------
                the key of the new added actor
        """
        key = 'T' + str(self._id)
        self.actors[key] = ActorEntity(self.text[char_span[0]:char_span[1]], char_span, lexical_head,
                                       actor_type)  # Hard-coded lexical head and type as 'Pronoun' and 'Other', resp., for now

        self._id += 1

        return key

    def _get_event_key(self, char_span, match_type="exact"):
        """
        Get the key of an event entity based on its character span on the full document text
        todo: Esta função está redundante com a _get_actor_key. Só muda a lista em que se procura

        @param char_span: The character span of the event in the text
        @return: The key of the event if it is found. None otherwise
        """

        if match_type == "exact":
            for key in self.events.keys():
                if self.events[key].character_span == char_span:
                    return str(key)
        elif match_type == "partial":
            for key in self.events.keys():
                aSpan = self.events[key].character_span
                if aSpan[0] <= char_span[0] <= aSpan[1]:
                    return str(key)
                elif aSpan[0] <= char_span[1] <= aSpan[1]:
                    return str(key)
        else:
            raise ValueError(
                f"Parameter match_type must be one of [exact, partial].\nInstead it was {match_type}")

        return None

    def _add_event(self, char_span):
        """
        Adds a new event to the narrative.

        @param char_span: tuple of characters (first_char, last_char) that delimit the event

        @return: The key of the new added event
        """
        key = 'T' + str(self._id)
        self.events[key] = EventEntity(
            self.text[char_span[0]:char_span[1]], char_span)

        self._id += 1

        return key

    def _add_time(self, char_span):
        """
        Adds a new time expression to the narrative.

        @param char_span: tuple of characters (first_char, last_char) that delimit the time expression

        @return: The key of the new added time expression
        """
        key = 'T' + str(self._id)

        # TODO: change default values for value and times_type parameters
        self.times[key] = TimeEntity(
            self.text[char_span[0]:char_span[1]], char_span, "","Time")

        self._id += 1

        return key

    def ISO_annotation(self):
        """
        Parameters
        ----------
                None

        Returns
        -------
                the ISO annotation in the .ann format
        """

        attribute_id = 1

        r = ""

        for actor_id in self.actors:
            actor = self.actors[actor_id]
            if '\n' in actor.text:
                continue
            # T1 ACTOR 0 22 O presidente de França
            r += (actor_id + '\t' + 'Participant' + ' ' + str(actor.character_span[0]) + ' ' + str(
                actor.character_span[1]) + '\t' + actor.text + '\n')

            # A1 Lexical_Head T1 Noun
            r += ('A' + str(attribute_id) + '\t' + 'Lexical_Head' +
                  ' ' + actor_id + ' ' + actor.lexical_head + '\n')
            attribute_id += 1

            # A2 Individuation T1 Individual
            r += ('A' + str(attribute_id) + '\t' + 'Individuation_Domain' +
                  ' ' + actor_id + ' ' + actor.individuation + '\n')
            attribute_id += 1

            # A3 Actor_Type T1 Per
            r += ('A' + str(attribute_id) + '\t' + 'Participant_Type_Domain' +
                  ' ' + actor_id + ' ' + actor.type + '\n')
            attribute_id += 1

            # A4 Involvement T1 1
            r += ('A' + str(attribute_id) + '\t' + 'Involvement' +
                  ' ' + actor_id + ' ' + actor.involvement + '\n')
            attribute_id += 1

        for time_id in self.times:
            time = self.times[time_id]

            # T26 TIME_X3 413 429 novembro de 2015
            r += (time_id + '\t' + 'Time' + ' ' + str(time.character_span[0]) + ' ' + str(
                time.character_span[1]) + '\t' + time.text + '\n')

            # A55 Time_Type T26 Date
            # r += ('A' + str(attribute_id) + '\t' + 'Time_Type' + ' ' + time_id + ' ' + time.type + '\n')
            # attribute_id += 1

            # 6 AnnotatorNotes T26 value=2015-11-XX  ?????
            # r += ('A' + str(attribute_id) + '\t' + 'Value' + ' ' + time_id + ' ' + time.value + '\n')
            # attribute_id += 1

            # A107 FunctionInDocument T4 Publication_Time
            # r += ('A' + str(attribute_id) + '\t' + 'FunctionInDocument' + ' ' + time_id + ' ' + time.temporal_function + '\n')
            r += ('A' + str(attribute_id) + '\t' + 'TemporalFunction' +
                  ' ' + time_id + ' ' + time.temporal_function + '\n')
            attribute_id += 1

        for event_id in self.events:

            event = self.events[event_id]
            if len(event.text.strip()) == 0:
                continue
            # T22 EVENT 312 328 is strengthening
            r += (event_id + '\t' + 'Event' + ' ' + str(event.character_span[0]) + ' ' + str(
                event.character_span[1]) + '\t' + event.text + '\n')

            # E9 EVENT:T22
            #r += (f"{event_id}\tEVENT:T{str(self._id)}\n")

            r += (f"A{attribute_id}\tClass {event_id} {event.event_class}\n")
            attribute_id += 1

            r += (f"A{attribute_id}\tTense {event_id} {event.tense}\n")
            attribute_id += 1

            r += (f"A{attribute_id}\tPolarity {event_id} {event.polarity}\n")
            attribute_id += 1

            #r += (f"A{attribute_id}\tFactuality {event_id} {event.factuality}\n")
            #import pdb
            # pdb.set_trace()
            attribute_id += 1

        for objectal_link_id in self.obj_links:
            obj_link = self.obj_links[objectal_link_id]

            # R34 OBJ_REL_objIdentity Arg1:T3 Arg2:T1
            #r += (objectal_link_id + '\t' + 'OBJ_REL_' + obj_link.type + ' ' + 'Arg1:' + obj_link.arg1 + ' ' + 'Arg2:' + obj_link.arg2 + '\n')
            r += (objectal_link_id + '\t' + 'OLINK_' + obj_link.type + ' ' +
                  'Arg1:' + obj_link.arg1 + ' ' + 'Arg2:' + obj_link.arg2 + '\n')

        for sem_link_id in self.sem_links:
            sem_link = self.sem_links[sem_link_id]

            # R35 SEMROLE_theme Arg1:E1 Arg2:T1
        #	r += (f"{sem_link_id}\tSEMROLE_{sem_link.type} Arg1:{sem_link.event} Arg2:{sem_link.actor}\n")
            r += (f"{sem_link_id}\tSRLINK_{sem_link.type} Arg1:{sem_link.event} Arg2:{sem_link.actor}\n")

        return r
